/*
 * Created on Jan 17, 2010
 * Created by Paul Gardner
 * 
 * Copyright 2010 Vuze, Inc.  All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License only.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


package com.aelitis.azureus.plugins.antithrottle;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.gudy.azureus2.plugins.Plugin;
import org.gudy.azureus2.plugins.PluginConfigListener;
import org.gudy.azureus2.plugins.PluginException;
import org.gudy.azureus2.plugins.PluginInterface;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadManagerStats;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.logging.LoggerChannelListener;
import org.gudy.azureus2.plugins.ui.UIManager;
import org.gudy.azureus2.plugins.ui.config.BooleanParameter;
import org.gudy.azureus2.plugins.ui.config.IntParameter;
import org.gudy.azureus2.plugins.ui.config.Parameter;
import org.gudy.azureus2.plugins.ui.model.BasicPluginConfigModel;
import org.gudy.azureus2.plugins.ui.model.BasicPluginViewModel;
import org.gudy.azureus2.plugins.utils.LocaleUtilities;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;


public class 
AntiThrottlePlugin 
	implements Plugin
{
	private PluginInterface		plugin_interface;
	private LoggerChannel		logger;
	
	public void 
	initialize(
		PluginInterface _plugin_interface )
	
		throws PluginException 
	{
		plugin_interface	= _plugin_interface;
		
		logger				= plugin_interface.getLogger().getTimeStampedChannel( "AntiThrottle" ); 


		LocaleUtilities loc_utils = plugin_interface.getUtilities().getLocaleUtilities();

		loc_utils.integrateLocalisedMessageBundle( "com.aelitis.azureus.plugins.antithrottle.internat.Messages" );

		UIManager	ui_manager	= plugin_interface.getUIManager();
		
		final BasicPluginViewModel	view_model = ui_manager.createBasicPluginViewModel( "azantithrottle.name" );

		view_model.getActivity().setVisible( false );
		view_model.getProgress().setVisible( false );
		
		logger.addListener(
				new LoggerChannelListener()
				{
					public void
					messageLogged(
						int		type,
						String	content )
					{
						view_model.getLogArea().appendText( content + "\n" );
					}
					
					public void
					messageLogged(
						String		str,
						Throwable	error )
					{
						if ( str.length() > 0 ){
							view_model.getLogArea().appendText( str + "\n" );
						}
						
						StringWriter sw = new StringWriter();
						
						PrintWriter	pw = new PrintWriter( sw );
						
						error.printStackTrace( pw );
						
						pw.flush();
						
						view_model.getLogArea().appendText( sw.toString() + "\n" );
					}
				});		
		
		BasicPluginConfigModel config_model = 
			ui_manager.createBasicPluginConfigModel( "azantithrottle.name" );

		view_model.setConfigSectionID( "azantithrottle.name" );
		
		final BooleanParameter 	ass_enable		= config_model.addBooleanParameter2( "azantithrottle.ass.enable", "azantithrottle.ass.enable", true );
		final IntParameter  	ass_min_speed 	= config_model.addIntParameter2( "azantithrottle.ass.min_down_speed", "azantithrottle.ass.min_down_speed", 25 );
		final IntParameter  	ass_trigger 	= config_model.addIntParameter2( "azantithrottle.ass.trigger_time", "azantithrottle.ass.trigger_time", 120 );
		final IntParameter  	ass_delay 		= config_model.addIntParameter2( "azantithrottle.ass.time_before_restart", "azantithrottle.ass.time_before_restart", 15 );
		final BooleanParameter 	ass_no_restart	= config_model.addBooleanParameter2( "azantithrottle.ass.norestart", "azantithrottle.ass.norestart", false );
		
		ass_enable.addEnabledOnSelection( ass_min_speed );
		ass_enable.addEnabledOnSelection( ass_trigger );
		ass_enable.addEnabledOnSelection( ass_delay );
		ass_enable.addEnabledOnSelection( ass_no_restart );
		
		config_model.createGroup(
			"azantithrottle.ass.name",
			new Parameter[]{ ass_enable, ass_min_speed, ass_trigger, ass_delay, ass_no_restart });
				
		plugin_interface.getUtilities().createTimer( "processor", true ).addPeriodicEvent(
			1000,
			new UTTimerEventPerformer()
			{
				private int	tick_count		= 0;
				
				private int	state			= 0;
				private int trigger_secs 	= 0;
				private int	wait_secs		= 0;
				
				private List<Download>	stopped_downloads  	= new ArrayList<Download>();
				private List<Download>	held_downloads  	= new ArrayList<Download>();
				
				public void 
				perform(
					UTTimerEvent event )
				{
					if ( !ass_enable.getValue()){
						
						state				= 0;
						trigger_secs 		= 0;
						wait_secs			= 0;
						
						stopped_downloads.clear();
						held_downloads.clear();
						
						return;
					}
					
					tick_count++;
					
					if ( tick_count % 30 == 0 ){
						
						for ( Iterator<Download> it=stopped_downloads.iterator();it.hasNext();){
	
							Download download = it.next();
							
							if ( download.isRemoved()){
								
								it.remove();
							}
						}
						
						for ( Iterator<Download> it=held_downloads.iterator();it.hasNext();){
	
							Download download = it.next();
							
							if ( download.isRemoved()){
								
								it.remove();
								
							}else if ( download.getLastScrapeResult().getSeedCount() > 0 ){
								
								it.remove();
								
								logger.log( "Download '" + download.getName() + "' has seeds - restarting" );
								
								try{
									download.restart();
									
								}catch( Throwable e ){
									
									logger.log( "Failed to restart '" + download.getName() + "'", e );
								}
							}
						}
					}
					
					if ( state == 0 ){
						
							// track download speed
						
						DownloadManagerStats stats = plugin_interface.getDownloadManager().getStats();
						
						int	download_speed = stats.getDataAndProtocolReceiveRate();
						
						if ((download_speed/1024) <= ass_min_speed.getValue()){
							
							trigger_secs++;
							
							if ( trigger_secs >= ass_trigger.getValue()){
								
								trigger_secs = 0;
								
								logger.log( "Min download speed of " + ass_min_speed.getValue() + " KB/sec not reached for " + ass_trigger.getValue() + " secs - stopping downloads" );
								
								state = 1;
							}
						}else{
							
							trigger_secs = 0;
						}
					}else if ( state == 1 ){
						
							// stop the downloads
						
						Download[] downloads = plugin_interface.getDownloadManager().getDownloads();
						
						for ( Download download: downloads ){
							
							if ( download.getState() == Download.ST_DOWNLOADING && !download.isComplete( false ) ){
								
								try{
									download.stop();
									
									stopped_downloads.add( download );
									
								}catch( Throwable e ){
									
									logger.log( "Failed to stop '" + download.getName() + "'", e );
								}
								
							}
						}
						
						if ( stopped_downloads.size() == 0 ){
						
							logger.log( "No downloads found to stop" );
							
						}else{
							
							logger.log( "Stopped " + stopped_downloads.size() + " downloads - waiting for " + ass_delay.getValue() + " secs before restarting" );
						}
						
						state = 2;
						
					}else if ( state == 2 ){
						
							// wait for restart
						
						wait_secs++;
						
						if ( wait_secs >= ass_delay.getValue()){
							
							wait_secs = 0;
							
							logger.log( "Restarting downloads" );
							
							state = 3;
						}
						
					}else{
						
						for ( Download download: stopped_downloads ){
							
							if ( ass_no_restart.getValue()){
								
								if ( download.getLastScrapeResult().getSeedCount() <= 0 ){
									
									if ( !held_downloads.contains( download )){
									
										held_downloads.add( download );
									
										logger.log( "Download '" + download.getName() + "' has no seeds - not restarting" );
									}
									
									continue;
								}
							}
							
							try{
								download.restart();
								
							}catch( Throwable e ){
								
								logger.log( "Failed to restart '" + download.getName() + "'", e );
							}
						}
						
						logger.log( "Restarted " + stopped_downloads.size() + " downloads" );
						
						stopped_downloads.clear();
						
							// retart
						
						state = 0;
					}

				}
			});
	}
}
